@@ -1,7 +1,7 @@
 /* regex.c: regular expression interface routines for the ed line editor. */
 /*  GNU ed - The GNU line editor.
     Copyright (C) 1993, 1994 Andrew Moore, Talke Studio
-    Copyright (C) 2006-2016 Antonio Diaz Diaz.
+    Copyright (C) 2006-2017 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -28,18 +28,14 @@
 #include "ed.h"
 
 
-static regex_t * global_pat = 0;
-static bool patlock = false;	/* if set, pattern not freed by get_compiled_pattern */
+static regex_t * subst_regex_ = 0;	/* regex of previous substitution */
 
-static char * stbuf = 0;	/* substitution template buffer */
-static int stbufsz = 0;		/* substitution template buffer size */
-static int stlen = 0;		/* substitution template length */
-
-static char * rbuf = 0;		/* replace_matching_text buffer */
-static int rbufsz = 0;		/* replace_matching_text buffer size */
+static char * rbuf = 0;		/* replacement buffer */
+static int rbufsz = 0;		/* replacement buffer size */
+static int rlen = 0;		/* replacement length */
 
 
-bool prev_pattern( void ) { return global_pat != 0; }
+bool subst_regex( void ) { return subst_regex_ != 0; }
 
 
 /* translate characters in a string */
@@ -107,62 +103,73 @@ static char * extract_pattern( const char ** const ibufpp, const char delimiter
   }
 
 
-/* return pointer to compiled pattern from command buffer */
-static regex_t * get_compiled_pattern( const char ** const ibufpp )
+/* return pointer to compiled regex from command buffer, or to previous
+   compiled regex if empty RE. return 0 if error */
+static regex_t * get_compiled_regex( const char ** const ibufpp,
+                                     const bool test_delimiter )
   {
+  static regex_t store[2];		/* space for two compiled regexes */
   static regex_t * exp = 0;
-  const char * exps;
+  const char * pat;
   const char delimiter = **ibufpp;
   int n;
 
   if( delimiter == ' ' )
     { set_error_msg( "Invalid pattern delimiter" ); return 0; }
-  if( delimiter == '\n' || *++*ibufpp == '\n' || **ibufpp == delimiter )
+  if( delimiter == '\n' || *++*ibufpp == delimiter ||
+      ( **ibufpp == '\n' && !test_delimiter ) )
     {
     if( !exp ) set_error_msg( "No previous pattern" );
     return exp;
     }
-  exps = extract_pattern( ibufpp, delimiter );
-  if( !exps ) return 0;
-  /* buffer alloc'd && not reserved */
-  if( exp && !patlock ) regfree( exp );
-  else
-    {
-    exp = (regex_t *) malloc( sizeof (regex_t) );
-    if( !exp )
-      {
-      show_strerror( 0, errno );
-      set_error_msg( "Memory exhausted" );
-      return 0;
-      }
-    }
-  patlock = false;
-  n = regcomp( exp, exps, 0 );
+  pat = extract_pattern( ibufpp, delimiter );
+  if( !pat ) return 0;
+  if( test_delimiter && delimiter != **ibufpp )
+    { set_error_msg( "Missing pattern delimiter" ); return 0; }
+  /* exp compiled && not copied */
+  if( exp && exp != subst_regex_ ) regfree( exp );
+  else exp = ( &store[0] != subst_regex_ ) ? &store[0] : &store[1];
+  n = regcomp( exp, pat, 0 );
   if( n )
     {
     char buf[80];
     regerror( n, exp, buf, sizeof buf );
     set_error_msg( buf );
-    free( exp );
     exp = 0;
     }
   return exp;
   }
 
 
-/* add line matching a pattern to the global-active list */
+bool set_subst_regex( const char ** const ibufpp )
+  {
+  regex_t * exp;
+
+  disable_interrupts();
+  exp = get_compiled_regex( ibufpp, true );
+  if( exp && exp != subst_regex_ )
+    {
+    if( subst_regex_ ) regfree( subst_regex_ );
+    subst_regex_ = exp;
+    }
+  enable_interrupts();
+  return ( exp ? true : false );
+  }
+
+
+/* add line matching a regular expression to the global-active list */
 bool build_active_list( const char ** const ibufpp, const int first_addr,
                         const int second_addr, const bool match )
   {
-  const regex_t * pat;
+  const regex_t * exp;
   const line_t * lp;
   int addr;
   const char delimiter = **ibufpp;
 
   if( delimiter == ' ' || delimiter == '\n' )
     { set_error_msg( "Invalid pattern delimiter" ); return false; }
-  pat = get_compiled_pattern( ibufpp );
-  if( !pat ) return false;
+  exp = get_compiled_regex( ibufpp, false );
+  if( !exp ) return false;
   if( **ibufpp == delimiter ) ++*ibufpp;
   clear_active_list();
   lp = search_line_node( first_addr );
@@ -171,73 +178,21 @@ bool build_active_list( const char ** const ibufpp, const int first_addr,
     char * const s = get_sbuf_line( lp );
     if( !s ) return false;
     if( isbinary() ) nul_to_newline( s, lp->len );
-    if( !regexec( pat, s, 0, 0, 0 ) == match && !set_active_node( lp ) )
+    if( match == !regexec( exp, s, 0, 0, 0 ) && !set_active_node( lp ) )
       return false;
     }
   return true;
   }
 
 
-/* return pointer to copy of substitution template in the command buffer */
-static char * extract_subst_template( const char ** const ibufpp,
-                                      const bool isglobal )
-  {
-  int i = 0, n = 0;
-  char c;
-  const char delimiter = **ibufpp;
-
-  ++*ibufpp;
-  if( **ibufpp == '%' && (*ibufpp)[1] == delimiter )
-    {
-    ++*ibufpp;
-    if( !stbuf ) set_error_msg( "No previous substitution" );
-    return stbuf;
-    }
-  while( **ibufpp != delimiter )
-    {
-    if( !resize_buffer( &stbuf, &stbufsz, i + 2 ) ) return 0;
-    c = stbuf[i++] = *(*ibufpp)++;
-    if( c == '\n' && **ibufpp == 0 ) { --i, --*ibufpp; break; }
-    if( c == '\\' && ( stbuf[i++] = *(*ibufpp)++ ) == '\n' && !isglobal )
-      {
-      while( ( *ibufpp = get_tty_line( &n ) ) &&
-             ( n == 0 || ( n > 0 && (*ibufpp)[n-1] != '\n' ) ) )
-        clearerr( stdin );
-      if( !*ibufpp ) return 0;
-      }
-    }
-  if( !resize_buffer( &stbuf, &stbufsz, i + 1 ) ) return 0;
-  stbuf[stlen = i] = 0;
-  return stbuf;
-  }
-
-
-/* extract substitution tail from the command buffer */
-bool extract_subst_tail( const char ** const ibufpp, int * const gflagsp,
-                         int * const snump, const bool isglobal )
-  {
-  const char delimiter = **ibufpp;
-
-  *gflagsp = *snump = 0;
-  if( delimiter == '\n' ) { stlen = 0; *gflagsp = GPR; return true; }
-  if( !extract_subst_template( ibufpp, isglobal ) ) return false;
-  if( **ibufpp == '\n' ) { *gflagsp = GPR; return true; }
-  if( **ibufpp == delimiter ) ++*ibufpp;
-  if( **ibufpp >= '1' && **ibufpp <= '9' )
-    return parse_int( snump, *ibufpp, ibufpp );
-  if( **ibufpp == 'g' ) { ++*ibufpp; *gflagsp = GSG; }
-  return true;
-  }
-
-
-/* return the address of the next line matching a pattern in a given
-   direction. wrap around begin/end of editor buffer if necessary */
+/* return the address of the next line matching a regular expression in a
+   given direction. wrap around begin/end of editor buffer if necessary */
 int next_matching_node_addr( const char ** const ibufpp, const bool forward )
   {
-  const regex_t * const pat = get_compiled_pattern( ibufpp );
+  const regex_t * const exp = get_compiled_regex( ibufpp, false );
   int addr = current_addr();
 
-  if( !pat ) return -1;
+  if( !exp ) return -1;
   do {
     addr = ( forward ? inc_addr( addr ) : dec_addr( addr ) );
     if( addr )
@@ -246,7 +201,7 @@ int next_matching_node_addr( const char ** const ibufpp, const bool forward )
       char * const s = get_sbuf_line( lp );
       if( !s ) return -1;
       if( isbinary() ) nul_to_newline( s, lp->len );
-      if( !regexec( pat, s, 0, 0, 0 ) ) return addr;
+      if( !regexec( exp, s, 0, 0, 0 ) ) return addr;
       }
     }
   while( addr != current_addr() );
@@ -255,134 +210,170 @@ int next_matching_node_addr( const char ** const ibufpp, const bool forward )
   }
 
 
-bool new_compiled_pattern( const char ** const ibufpp )
+/* extract substitution replacement from the command buffer */
+bool extract_replacement( const char ** const ibufpp, const bool isglobal )
   {
-  regex_t * tpat;
+  static char * buf = 0;		/* temporary buffer */
+  static int bufsz = 0;
+  int i = 0;
+  const char delimiter = **ibufpp;
 
-  disable_interrupts();
-  tpat = get_compiled_pattern( ibufpp );
-  if( tpat && tpat != global_pat )
+  if( delimiter == '\n' )
+    { set_error_msg( "Missing pattern delimiter" ); return false; }
+  ++*ibufpp;
+  if( **ibufpp == '%' && ( (*ibufpp)[1] == delimiter || (*ibufpp)[1] == '\n' ) )
     {
-    if( global_pat ) { regfree( global_pat ); free( global_pat ); }
-    global_pat = tpat;
-    patlock = true;		/* reserve pattern */
+    ++*ibufpp;
+    if( !rbuf ) { set_error_msg( "No previous substitution" ); return false; }
+    return true;
     }
+  while( **ibufpp != delimiter )
+    {
+    if( **ibufpp == '\n' )
+      {
+      if( isglobal && (*ibufpp)[1] != 0 )
+        { set_error_msg( "Invalid newline substitution" ); return false; }
+      break;
+      }
+    if( !resize_buffer( &buf, &bufsz, i + 2 ) ) return false;
+    if( ( buf[i++] = *(*ibufpp)++ ) == '\\' &&
+        ( buf[i++] = *(*ibufpp)++ ) == '\n' && !isglobal )
+      {
+      /* not reached if isglobal; in command-list, newlines are unescaped */
+      int size = 0;
+      *ibufpp = get_stdin_line( &size );
+      if( !*ibufpp ) return false;			/* error */
+      if( size <= 0 ) return false;			/* EOF */
+      }
+    }
+  /* make sure that buf gets allocated if empty replacement */
+  if( !resize_buffer( &buf, &bufsz, i + 1 ) ) return false;
+  buf[i] = 0;
+  disable_interrupts();
+  { char * p = buf; buf = rbuf; rbuf = p;		/* swap buffers */
+    rlen = i; i = bufsz; bufsz = rbufsz; rbufsz = i; }
   enable_interrupts();
-  return ( tpat ? true : false );
+  return true;
   }
 
 
-/* modify text according to a substitution template; return offset to
-   end of modified text */
-static int apply_subst_template( const char * const boln,
+/* Produce replacement text from matched text and replacement template.
+   Return new offset to end of replacement text, or -1 if error. */
+static int replace_matched_text( char ** txtbufp, int * const txtbufszp,
+                                 const char * const txt,
                                  const regmatch_t * const rm, int offset,
                                  const int re_nsub )
   {
-  const char * sub = stbuf;
+  const char * sub = rbuf;
 
-  for( ; sub - stbuf < stlen; ++sub )
+  for( ; sub - rbuf < rlen; ++sub )
     {
     int n;
     if( *sub == '&' )
       {
       int j = rm[0].rm_so; int k = rm[0].rm_eo;
-      if( !resize_buffer( &rbuf, &rbufsz, offset + k - j ) ) return -1;
-      while( j < k ) rbuf[offset++] = boln[j++];
+      if( !resize_buffer( txtbufp, txtbufszp, offset + k - j ) ) return -1;
+      while( j < k ) (*txtbufp)[offset++] = txt[j++];
       }
     else if( *sub == '\\' && *++sub >= '1' && *sub <= '9' &&
              ( n = *sub - '0' ) <= re_nsub )
       {
       int j = rm[n].rm_so; int k = rm[n].rm_eo;
-      if( !resize_buffer( &rbuf, &rbufsz, offset + k - j ) ) return -1;
-      while( j < k ) rbuf[offset++] = boln[j++];
+      if( !resize_buffer( txtbufp, txtbufszp, offset + k - j ) ) return -1;
+      while( j < k ) (*txtbufp)[offset++] = txt[j++];
       }
     else
       {
-      if( !resize_buffer( &rbuf, &rbufsz, offset + 1 ) ) return -1;
-      rbuf[offset++] = *sub;
+      if( !resize_buffer( txtbufp, txtbufszp, offset + 1 ) ) return -1;
+      (*txtbufp)[offset++] = *sub;
       }
     }
-  if( !resize_buffer( &rbuf, &rbufsz, offset + 1 ) ) return -1;
-  rbuf[offset] = 0;
+  if( !resize_buffer( txtbufp, txtbufszp, offset + 1 ) ) return -1;
+  (*txtbufp)[offset] = 0;
   return offset;
   }
 
 
-/* replace text matched by a pattern according to a substitution
-   template; return size of the modified text */
-static int replace_matching_text( const line_t * const lp, const int gflags,
-                                  const int snum )
+/* Produce new text with one or all matches replaced in a line.
+   Return size of the new line text, 0 if no change, -1 if error */
+static int line_replace( char ** txtbufp, int * const txtbufszp,
+                         const line_t * const lp, const int snum )
   {
   enum { se_max = 30 };	/* max subexpressions in a regular expression */
   regmatch_t rm[se_max];
   char * txt = get_sbuf_line( lp );
   const char * eot;
   int i = 0, offset = 0;
+  const bool global = ( snum <= 0 );
   bool changed = false;
 
   if( !txt ) return -1;
   if( isbinary() ) nul_to_newline( txt, lp->len );
   eot = txt + lp->len;
-  if( !regexec( global_pat, txt, se_max, rm, 0 ) )
+  if( !regexec( subst_regex_, txt, se_max, rm, 0 ) )
     {
     int matchno = 0;
     do {
-      if( !snum || snum == ++matchno )
+      if( global || snum == ++matchno )
         {
         changed = true; i = rm[0].rm_so;
-        if( !resize_buffer( &rbuf, &rbufsz, offset + i ) ) return -1;
+        if( !resize_buffer( txtbufp, txtbufszp, offset + i ) ) return -1;
         if( isbinary() ) newline_to_nul( txt, rm[0].rm_eo );
-        memcpy( rbuf + offset, txt, i ); offset += i;
-        offset = apply_subst_template( txt, rm, offset, global_pat->re_nsub );
+        memcpy( *txtbufp + offset, txt, i ); offset += i;
+        offset = replace_matched_text( txtbufp, txtbufszp, txt, rm, offset,
+                                       subst_regex_->re_nsub );
         if( offset < 0 ) return -1;
         }
       else
         {
         i = rm[0].rm_eo;
-        if( !resize_buffer( &rbuf, &rbufsz, offset + i ) ) return -1;
+        if( !resize_buffer( txtbufp, txtbufszp, offset + i ) ) return -1;
         if( isbinary() ) newline_to_nul( txt, i );
-        memcpy( rbuf + offset, txt, i ); offset += i;
+        memcpy( *txtbufp + offset, txt, i ); offset += i;
         }
       txt += rm[0].rm_eo;
       }
-    while( *txt && ( !changed || ( ( gflags & GSG ) && rm[0].rm_eo ) ) &&
-           !regexec( global_pat, txt, se_max, rm, REG_NOTBOL ) );
+    while( *txt && ( !changed || ( global && rm[0].rm_eo ) ) &&
+           !regexec( subst_regex_, txt, se_max, rm, REG_NOTBOL ) );
     i = eot - txt;
-    if( !resize_buffer( &rbuf, &rbufsz, offset + i + 2 ) ) return -1;
-    if( i > 0 && !rm[0].rm_eo && ( gflags & GSG ) )
+    if( !resize_buffer( txtbufp, txtbufszp, offset + i + 2 ) ) return -1;
+    if( global && i > 0 && !rm[0].rm_eo )
       { set_error_msg( "Infinite substitution loop" ); return -1; }
     if( isbinary() ) newline_to_nul( txt, i );
-    memcpy( rbuf + offset, txt, i );
-    memcpy( rbuf + offset + i, "\n", 2 );
+    memcpy( *txtbufp + offset, txt, i );		/* tail copy */
+    memcpy( *txtbufp + offset + i, "\n", 2 );
     }
   return ( changed ? offset + i + 1 : 0 );
   }
 
 
-/* for each line in a range, change text matching a pattern according to
-   a substitution template; return false if error */
+/* for each line in a range, change text matching a regular expression
+   according to a substitution template (replacement); return false if error */
 bool search_and_replace( const int first_addr, const int second_addr,
-                         const int gflags, const int snum, const bool isglobal )
+                         const int snum, const bool isglobal )
   {
+  static char * txtbuf = 0;		/* new text of line buffer */
+  static int txtbufsz = 0;		/* new text of line buffer size */
+  int addr = first_addr;
   int lc;
   bool match_found = false;
 
-  set_current_addr( first_addr - 1 );
-  for( lc = 0; lc <= second_addr - first_addr; ++lc )
+  for( lc = 0; lc <= second_addr - first_addr; ++lc, ++addr )
     {
-    const line_t * const lp = search_line_node( inc_current_addr() );
-    const int size = replace_matching_text( lp, gflags, snum );
+    const line_t * const lp = search_line_node( addr );
+    const int size = line_replace( &txtbuf, &txtbufsz, lp, snum );
     if( size < 0 ) return false;
     if( size )
       {
-      const char * txt = rbuf;
-      const char * const eot = rbuf + size;
+      const char * txt = txtbuf;
+      const char * const eot = txtbuf + size;
       undo_t * up = 0;
       disable_interrupts();
-      if( !delete_lines( current_addr(), current_addr(), isglobal ) )
+      if( !delete_lines( addr, addr, isglobal ) )
         { enable_interrupts(); return false; }
+      set_current_addr( addr - 1 );
       do {
-        txt = put_sbuf_line( txt, size, current_addr() );
+        txt = put_sbuf_line( txt, eot - txt );
         if( !txt ) { enable_interrupts(); return false; }
         if( up ) up->tail = search_line_node( current_addr() );
         else
@@ -393,10 +384,11 @@ bool search_and_replace( const int first_addr, const int second_addr,
         }
       while( txt != eot );
       enable_interrupts();
+      addr = current_addr();
       match_found = true;
       }
     }
-  if( !match_found && !( gflags & GLB ) )
+  if( !match_found && !isglobal )
     { set_error_msg( "No match" ); return false; }
   return true;
   }
